unit MainUnit;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, ExtCtrls,
  TintProgresses,
  GenIRQThreads;

type
  TCardData = record
    CardNum, IRQCount: Integer;
    GenIRQThread: TGenIRQThread;
  end;

  TMainForm = class(TForm)
    LogMemo: TMemo;
    ButtonPanel: TPanel;
    TestButton: TButton;
    SaveEm: TSaveDialog;
    procedure FormCreate(Sender: TObject);
    procedure TestButtonClick(Sender: TObject);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
  protected
    Prog: TTintProgress;
    bCancel: Boolean;

    procedure TestInner;
    { Private declarations }
  public
    CardData: array of TCardData;
    { Public declarations }
  end;

var
  MainForm: TMainForm;

implementation

{$R *.DFM}

uses
  ShlObj, AIOWDM;

function PollForIRQ(CardNum: Integer): LongWord; cdecl; external 'AIOWDM.dll';

function PluralS(Count: Integer): AnsiString;
begin
  if Count = 1 then
    Result := ''
  else
    Result := 's'
  ;
end;

procedure LoadDefaultFont(IntoFont: TFont);
var
  FontSize: Integer;
  NCM: TNonClientMetrics;

begin
  NCM.cbSize := SizeOf(NCM);
  if SystemParametersInfo(SPI_GETNONCLIENTMETRICS, SizeOf(NCM), @NCM, 0) then begin
    FontSize := Round(-3/4 * NCM.lfMessageFont.lfHeight);
    IntoFont.Name := NCM.lfMessageFont.lfFaceName;
  end
  else begin
    FontSize := Abs(IntoFont.Size);
    IntoFont.Name := 'Tahoma';
  end;
  IntoFont.PixelsPerInch := 72;
  IntoFont.Size := FontSize;
end;

function GetCSIDLPath(CSIDL: Integer): String;
var
  I: Integer;

begin
  SetLength(Result, MAX_PATH + 1);
  if SHGetSpecialFolderPath(Application.Handle, @Result[1], CSIDL, False) then begin
    for I := 1 to Length(Result) do
      if Result[I] = #0 then begin
        SetLength(Result, I - 1);
        Break;
      end
    ;
  end
  else begin
    Result := '';
  end;
end;

type
  TTestPass = (tpAllEnable, tpIndividualEnable, tpReceptive);
  TEdgeType = (etStillLow, etRising, etFalling, etStillHigh);
  TResultEntry = record
    Tot, Wt: Integer;
  end;

const
  CodeFromTestPass: array[TTestPass] of AnsiString = ('A', 'I', 'R');
  CodeFromEdgeType: array[TEdgeType] of AnsiString = ('L', 'R', 'F', 'H');

function TestBit(TestAndCard, iBit: Integer): Boolean;
begin
  //Result := (TestAndCard and (1 shl iBit)) <> 0;
  {
  iBit := iBit * 2;
  Result :=
    ((TestAndCard and (1 shl iBit)) <> 0)
  xor
    ((TestAndCard and (2 shl iBit)) <> 0)
  ;
  }
  TestAndCard := TestAndCard + Length(MainForm.CardData) * iBit;
  {
  Result :=
    ((TestAndCard and 1) <> 0)
  xor
    ((TestAndCard and 2) <> 0)
  ;
  }
  Result := False;
  while TestAndCard <> 0 do begin
    Result := Result xor ((TestAndCard and 1) <> 0);
    TestAndCard := TestAndCard shr 1;
  end;
end;

function IncidenceStr(Entry: TResultEntry): AnsiString;
begin
  with Entry do begin
    //Result := IntToStr(Tot) + '/' + IntToStr(Wt); {
    if Tot = 0 then begin
      if Wt = 0 then
        Result := '?'
      else
        Result := '0'
      ;
    end
    else if Tot = Wt then
      Result := '1'
    else begin
      Result := FloatToStr(Tot / Wt);
      if Result = '1' then
        Result := '0.99999999999999'
      else if Result = '0' then
        Result := '0.00000000000001'
      ;
    end;
    //}
  end;
end;

procedure TMainForm.FormCreate(Sender: TObject);
var
  NumCards, iCardNum, I: Integer;
  DeviceID: LongWord;

begin
  LoadDefaultFont(Font);
  SaveEm.InitialDir := GetCSIDLPath(CSIDL_DESKTOP);
  
  NumCards := GetNumCards;
  LogMemo.Lines.Add(IntToStr(NumCards) + ' total card' + PluralS(Length(CardData)) + ' detected.');
  for iCardNum := 0 to NumCards-1 do begin
    if QueryCardInfo(iCardNum, @DeviceID, nil, nil, nil) = 0 then Continue;
    if DeviceID <> $0C78 then Continue;

    I := Length(CardData);
    SetLength(CardData, I+1);
    CardData[I].CardNum := iCardNum;
    CardData[I].IRQCount := 0;
  end;

  LogMemo.Lines.Add(IntToStr(Length(CardData)) + ' PCI-DIO-120 card' + PluralS(Length(CardData)) + ' detected.');
  LogMemo.Lines.Add('');
  if Length(CardData) = 0 then begin
    ButtonPanel.Visible := False;
  end
  else begin
    LogMemo.Lines.Add('Please unplug any connectors plugged into the card' + PluralS(Length(CardData)) + ', and then click Test.');
    LogMemo.Lines.Add('');
    LogMemo.Lines.Add('(This test produces lots of outputs which may interfere with external equipment, and inputs from external equipment may interfere with this test.)');
    LogMemo.Lines.Add('');
    Prog := TTintProgress.Create(Self);
    Prog.Align := alBottom;
    Prog.Parent := Self;
  end;
end;

procedure TMainForm.TestButtonClick(Sender: TObject);
var
  iCard, iPPI: Integer;

begin
  TestButton.Enabled := False;

  LogMemo.Lines.Clear;
  Prog.Position := 0;

  bCancel := False;
  
  for iCard := 0 to High(CardData) do with CardData[iCard] do begin
    RelOutPortB(CardNum, $1E, 0);
    IRQCount := 0;
    GenIRQThread := TGenIRQThread.Create(True);
    GenIRQThread.iCard := iCard;
    GenIRQThread.Resume;

    {
    for iPPI := 0 to 4 do
      RelOutPortB(CardNum, iPPI*4 + 2, $FF)
    ;

    RelOutPortB(CardNum, $1F, 0);
    }

    for iPPI := 0 to 4 do begin
      RelOutPortB(CardNum, iPPI*4 + 3, $80);
      RelOutPort (CardNum, iPPI*4 + 0, $FF);
      RelOutPortB(CardNum, iPPI*4 + 2, $7F);
    end;
    RelOutPortB(CardNum, $1F, 0);
    Sleep(1);
    PollForIRQ(CardNum); //Eat any IRQ from this config.
  end;

  TestInner;

  for iCard := 0 to High(CardData) do
    for iPPI := 0 to 4 do
      RelOutPortB(CardData[iCard].CardNum, iPPI*4 + 3, $9B)
  ;

  for iCard := 0 to High(CardData) do
    AbortRequest(CardData[iCard].CardNum)
  ;
  for iCard := 0 to High(CardData) do
    FreeAndNil(CardData[iCard].GenIRQThread)
  ;

  LogMemo.Lines.Add('Done.');

  TestButton.Enabled := True;
end;

procedure TMainForm.TestInner;
var
  iTest, iCard, iPPI: Integer;
  TestPass: TTestPass;
  TestsPerPass, Passes: Integer;
  Edge: TEdgeType;
  CData: Byte;
  ResultTable: array of array[TTestPass, 0..4, TEdgeType] of TResultEntry;
  IRQCount: Integer;
  Buf: AnsiString;
  LogFile: TStringList;

begin
  SetLength(ResultTable, Length(CardData));
  for iCard := 0 to High(CardData) do
    for TestPass := Low(TestPass) to High(TestPass) do
      for iPPI := 0 to 4 do
        for Edge := Low(Edge) to High(Edge) do
          with ResultTable[iCard, TestPass, iPPI, Edge] do begin
            Tot := 0;
            Wt := 0;
          end
  ;
  Passes := Ord(High(TestPass)) - Ord(Low(TestPass)) + 1;
  TestsPerPass := 63 + Length(CardData) * 6;
  Prog.Min := 0;
  Prog.Max := TestsPerPass * Passes;

  for TestPass := Low(TestPass) to High(TestPass) do begin
    LogMemo.Lines.Add('Pass ' + CodeFromTestPass[TestPass] + '');
    for iCard := 0 to High(CardData) do begin
      for iPPI := 0 to 4 do
        RelOutPortB(CardData[iCard].CardNum, iPPI*4 + 2, $FF)
      ;
      InterlockedExchange(CardData[iCard].IRQCount, 0);
    end;
    for iTest := 0 to TestsPerPass-1 do begin
      if bCancel then begin
        LogMemo.Lines.Add('Test aborted by user.');
        Exit;
      end;
      Prog.Position := Prog.Position + 1;
      for iCard := 0 to High(CardData) do begin
        if CardData[iCard].GenIRQThread.LastStatus <> ERROR_SUCCESS then begin
          LogMemo.Lines.Add('Error ' + IntToStr(CardData[iCard].GenIRQThread.LastStatus) + ' from WaitForIRQ(' + IntToStr(CardData[iCard].CardNum) + ')! Aborting test.');
          Exit;
        end;
        for iPPI := 0 to 4 do begin
          InterlockedExchange(CardData[iCard].IRQCount, 0);

          CData := $FF;
          if TestPass < tpReceptive then
            CData := CData and not (1 shl 7)
          ;
          if not TestBit(iTest + iCard, iPPI) then
            CData := CData and not (1 shl 3)
          ;
          RelOutPortB(CardData[iCard].CardNum, iPPI*4 + 2, CData);

          Sleep(1);

          Edge := Low(Edge);
          if TestBit(iTest + iCard, iPPI) then
            Inc(Edge)
          ;
          if (iTest = 0) or TestBit(iTest + iCard - 1, iPPI) then
            Inc(Edge, 2)
          ;

          IRQCount := InterlockedExchange(CardData[iCard].IRQCount, 0);
          if iTest <> 0 then
            with ResultTable[iCard, TestPass, iPPI, Edge] do begin
              if IRQCount <> 0 then
                Inc(Tot)
              ;
              Inc(Wt);
            end
          ;

          if TestPass > tpAllEnable then begin
            CData := CData or (1 shl 7);
            RelOutPortB(CardData[iCard].CardNum, iPPI*4 + 2, CData);
          end;
        end;
      end;
      Application.ProcessMessages;
    end;
  end;
  Prog.Position := Prog.Max;

  LogFile := TStringList.Create;

  Buf := 'Card:PPI';
  for Edge := Low(Edge) to High(Edge) do
    for TestPass := Low(TestPass) to High(TestPass) do
      Buf := Buf + (#9 + CodeFromEdgeType[Edge]) + CodeFromTestPass[TestPass]
  ;
  LogMemo.Lines.Add(Buf);
  LogFile.Add(Buf);
  for iCard := 0 to High(CardData) do
    for iPPI := 0 to 4 do begin
      //LogMemo.Lines.Add();
      Buf := IntToStr(iCard) + ':' + IntToStr(iPPI);
      for Edge := Low(Edge) to High(Edge) do
        for TestPass := Low(TestPass) to High(TestPass) do
          Buf := Buf + (#9 + IncidenceStr(ResultTable[iCard, TestPass, iPPI, Edge]))
      ;
      LogMemo.Lines.Add(Buf);
      LogFile.Add(Buf);
    end
  ;
  SaveEm.FileName := 'Log.tsv';
  if SaveEm.Execute then begin
    SaveEm.InitialDir := ExtractFilePath(SaveEm.FileName);
    LogFile.SaveToFile(SaveEm.FileName);
  end;
  LogFile.Free;
end;

procedure TMainForm.FormClose(Sender: TObject; var Action: TCloseAction);
begin
  bCancel := True;
end;

end.
